# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements transformer layers for MaX-DeepLab [1] and kMaX-DeepLab [2].

Dual-path transformer introduces a global memory path in addition to a CNN path,
allowing bi-directional communication with any CNN layers.

k-means cross-attention adopts a cluster-wise argmax instead of spatial-wise
softmax, which aligns to k-means clustering algorithm.

[1] MaX-DeepLab: End-to-End Panoptic Segmentation with Mask Transformers,
    CVPR 2021.
      Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.

[2] k-means Mask Transformer, ECCV 2022.
      Qihang Yu, Huiyu Wang, Siyuan Qiao, Maxwell Collins, Yukun Zhu,
      Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
"""

import tensorflow as tf

from deeplab2 import common
from deeplab2.model import utils
from deeplab2.model.layers import activations
from deeplab2.model.layers import convolutions


class AttentionOperation(tf.keras.layers.Layer):
    """Computes standard 1D multi-head attention with query, key, and value."""

    def __init__(self,
                 name,
                 activation,
                 transformer_activation,
                 bn_layer=tf.keras.layers.BatchNormalization):
        """Initializes an AttentionOperation layer.

        Args:
          name: A string, the name of this layer.
          activation: A string, type of activation function to apply.
          transformer_activation: A string, type of activation function for
            self-attention. Support 'sigmoid' and 'softmax'.
          bn_layer: An optional tf.keras.layers.Layer that computes the
            normalization (default: tf.keras.layers.BatchNormalization).
        """
        super(AttentionOperation, self).__init__(name=name)
        # batch_norm_similarity has shape [batch, num_heads, num_query, num_key],
        # where num_query and num_key usually equals to height or width or length,
        # i.e., spatial dimensions, so batch norm is applied to axis=1 only.
        self._batch_norm_similarity = bn_layer(axis=1, name='batch_norm_similarity')
        # batch_norm_retrieved_value is done on shape [batch, num_heads, length,
        # value_channels], which will be reshaped to the output shape [batch,
        # length, value_channels * num_heads], so we apply batch norm on the
        # effective channel dimension -- value_channels * num_heads.
        self._batch_norm_retrieved_value = bn_layer(
            axis=[1, 3], name='batch_norm_retrieved_value')
        self._activation_fn = activations.get_activation(activation)
        self._transformer_activation_fn = activations.get_activation(
            transformer_activation)

    def call(self, inputs, training=False):
        """Performs an AttentionOperation.

        Args:
          inputs: A tuple of (query, key, value), where query is [batch, num_head,
            query_length, channels] tensor, key is a [batch, num_head, key_length,
            channels] tensor, and value is a [batch, key_length, num_head,
            value_channels] tensor.
          training: A boolean, whether the model is in training mode.

        Returns:
          output: A [batch, query_length, num_head * value_channels] tensor, the
            retrieved value.
        """
        # Decode query, key, and value from inputs.
        query, key, value = inputs
        # Compute attention similarity.
        similarity_logits = tf.einsum('bhld,bhmd->bhlm', query, key)
        similarity_logits = self._batch_norm_similarity(
            similarity_logits, training=training)
        # Apply a transformer attention activation function, e.g. softmax.
        attention_weights = self._transformer_activation_fn(similarity_logits)
        # Retrieve the value content.
        retrieved_value = tf.einsum(
            'bhlm,bmhd->bhld', attention_weights, value)
        retrieved_value = self._batch_norm_retrieved_value(
            retrieved_value, training=training)
        retrieved_value = self._activation_fn(retrieved_value)
        # Reshape the output.
        return utils.transpose_and_reshape_for_attention_operation(
            retrieved_value)


class DualPathTransformerLayer(tf.keras.layers.Layer):
    """Applies a transformer layer, as proposed in MaX-DeepLab models [1,2].

    Dual-path transformer layer takes a pixel space input and a memory space
    input, and performs memory2pixel attention, pixel2memory attention, and
    memory2memory self-attention. Note that the pixel2pixel self-attention or
    convolution in the pixel space is implemented in axial_layers.py and
    axial_blocks.py. Thus, the pixel2pixel operation is not included in this
    DualPathTransformerLayer implementation. Please use this class together with
    a residual block with axial-attention, global-attention, or convolution in
    order to construct the full dual path transformer in the paper.

    The flag "use_kmeans_cross_attention" enables k-means cross-attention, which
    regards the memory (object query) as cluster center, and updates them in a
    k-means clustering manner.

    [1] MaX-DeepLab: End-to-End Panoptic Segmentation with Mask Transformers,
        CVPR 2021.
          Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.

    [2] k-means Mask Transformer, ECCV 2022.
        Qihang Yu, Huiyu Wang, Siyuan Qiao, Maxwell Collins, Yukun Zhu,
        Hartwig Adam, Alan Yuille, Liang-Chieh Chen.
    """

    def __init__(self,
                 name='dual_path_transformer_layer',
                 activation='relu',
                 filters=128,
                 num_heads=8,
                 bottleneck_expansion=2,
                 key_expansion=1,
                 value_expansion=2,
                 feed_forward_network_channels=2048,
                 use_memory_self_attention=True,
                 use_memory2pixel_feedback_attention=True,
                 use_pixel2memory_feedback_attention=True,
                 use_kmeans_cross_attention=False,
                 transformer_activation='softmax',
                 bn_layer=tf.keras.layers.BatchNormalization,
                 conv_kernel_weight_decay=0.0,
                 auxiliary_predictor_func=None):
        """Initializes a DualPathTransformerLayer.

        This function implements a dual path transformer layer between a pixel space
        and a memory space, as described in the MaX-DeepLab paper. In this dual path
        transformer, the memory2pixel cross attention and the memory self-attention
        share a single activation, e.g. softmax.

        The flag "use_kmeans_cross_attention" enables k-means cross-attention
        proposed in the kMaX-DeepLab paper, which regards the memory (object query)
        as cluster center, and updates them in a k-means clustering manner.

        Reference:
          MaX-DeepLab: "End-to-End Panoptic Segmentation with Mask Transformers",
            CVPR 2021. https://arxiv.org/abs/2012.00759
              Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.

          k-means Mask Transformer, ECCV 2022.
            Qihang Yu, Huiyu Wang, Siyuan Qiao, Maxwell Collins, Yukun Zhu,
            Hartwig Adam, Alan Yuille, Liang-Chieh Chen.

        Args:
          name: A string, the name of this dual path transformer layer.
          activation: A string, type of activation function to apply.
          filters: An integer, the base number of channels for the layer.
          num_heads: An integer, the number of heads in multi-head attention.
          bottleneck_expansion: A float, the channel expansion ratio for the
            bottleneck.
          key_expansion: A float, the channel expansion ratio for keys.
          value_expansion: A float, the channel expansion ratio for values.
          feed_forward_network_channels: An integer, the number of channels for the
            feed_forward_network. Zero means no feed_forward_network will be
            applied.
          use_memory_self_attention: A boolean, whether to apply the memory space
            self-attention.
          use_memory2pixel_feedback_attention: A boolean, whether to apply the
            memory2pixel feedback attention.
          use_pixel2memory_feedback_attention: A boolean, whether to apply the
            pixel2memory feedback attention.
          use_kmeans_cross_attention: A boolean, whether to apply the kmeans
            cross-attention.
          transformer_activation: A string, type of activation function for
            self-attention. Support 'sigmoid' and 'softmax'.
          bn_layer: A tf.keras.layers.Layer that computes the normalization
            (default: tf.keras.layers.BatchNormalization).
          conv_kernel_weight_decay: A float, the weight decay for convolution
            kernels.
          auxiliary_predictor_func: A callable function that returns an
            initialization of auxiliary predictor.

        Raises:
          ValueError: If filters * key_expansion is not divisible by num_heads.
          ValueError: If filters * value_expansion is not divisible by num_heads.
          ValueError: If both use_memory2pixel_feedback_attention and
            use_kmeans_cross_attention are False.
          ValueError: If use_kmeans_cross_attention is True but
            auxiliary_predictor_func is None.
        """
        super(DualPathTransformerLayer, self).__init__(name=name)

        bottleneck_channels = int(round(filters * bottleneck_expansion))
        total_key_depth = int(round(filters * key_expansion))
        total_value_depth = int(round(filters * value_expansion))

        if total_key_depth % num_heads:
            raise ValueError('Total_key_depth should be divisible by num_heads.')

        if total_value_depth % num_heads:
            raise ValueError('Total_value_depth should be divisible by num_heads.')

        if not (use_memory2pixel_feedback_attention or use_kmeans_cross_attention):
            raise ValueError('At least one of use_memory2pixel_feedback_attention or'
                             ' use_kmeans_cross_attention needs to be enabled.')

        # Compute query key value with one convolution and a batch norm layer. The
        # initialization std is standard transformer initialization (without batch
        # norm), as used in SASA and ViT. In our case, we use batch norm by default,
        # so it does not require careful tuning. If one wants to remove all batch
        # norms in axial attention, this standard initialization should still be
        # good, but a more careful initialization is encouraged.
        initialization_std = bottleneck_channels ** -0.5

        self._memory_conv1_bn_act = convolutions.Conv1D(
            bottleneck_channels, 'memory_conv1_bn_act',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=activation,
            conv_kernel_weight_decay=conv_kernel_weight_decay)

        self._pixel_conv1_bn_act = convolutions.Conv1D(
            bottleneck_channels, 'pixel_conv1_bn_act',
            use_bias=False,
            use_bn=True,
            bn_layer=bn_layer,
            activation=activation,
            conv_kernel_weight_decay=conv_kernel_weight_decay)

        # We always compute the query for memory space, since it gathers information
        # from the pixel space and thus cannot be removed. We compute the key and
        # value for memory space only when they are necessary (i.e. either
        # use_memory_self_attention or use_pixel2memory_feedback_attention).
        if use_memory_self_attention or use_pixel2memory_feedback_attention:
            self._memory_qkv_conv_bn = convolutions.Conv1D(
                total_key_depth * 2 + total_value_depth, 'memory_qkv_conv_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=initialization_std))
        elif use_memory2pixel_feedback_attention:
            # Compute memory query only if memory key and value are not used.
            self._memory_query_conv_bn = convolutions.Conv1D(
                total_key_depth,
                'memory_query_conv_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=initialization_std))

        # For the pixel space, we always compute the key and value, since they
        # provide information for the memory space and thus cannot be removed. We
        # compute the query for pixel space only when it is necessary (i.e.
        # use_pixel2memory_feedback_attention is True).
        if use_pixel2memory_feedback_attention:
            self._pixel_qkv_conv_bn = convolutions.Conv1D(
                total_key_depth * 2 + total_value_depth, 'pixel_qkv_conv_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=initialization_std))
        elif use_memory2pixel_feedback_attention:
            self._pixel_kv_conv_bn = convolutions.Conv1D(
                total_key_depth + total_value_depth, 'pixel_kv_conv_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=initialization_std))
        else:
            # In this case, only k-means cross attention is enabled, and thus we only
            # need the value of pixel to update memory.
            self._pixel_v_conv_bn = convolutions.Conv1D(
                total_value_depth, 'pixel_v_conv_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=bn_layer,
                activation='none',
                conv_kernel_weight_decay=conv_kernel_weight_decay,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=initialization_std))

        if use_memory2pixel_feedback_attention or use_memory_self_attention:
            self._memory_attention = AttentionOperation(
                'memory_attention', activation, transformer_activation,
                bn_layer=bn_layer)
        if use_pixel2memory_feedback_attention:
            self._pixel_attention = AttentionOperation(
                'pixel_attention', activation, transformer_activation,
                bn_layer=bn_layer)

        if use_kmeans_cross_attention:
            # If kmeans cross-attention is used, we perform an auxiliary prediction.
            if auxiliary_predictor_func is None:
                raise ValueError('auxiliary_predictor_func should not be None when'
                                 ' using kmeans cross attention.')
            self._auxiliary_clustering_predictor = auxiliary_predictor_func()

        self._use_memory_self_attention = use_memory_self_attention
        self._use_memory2pixel_feedback_attention = (
            use_memory2pixel_feedback_attention)
        self._use_pixel2memory_feedback_attention = (
            use_pixel2memory_feedback_attention)
        self._use_kmeans_cross_attention = use_kmeans_cross_attention
        self._bottleneck_channels = bottleneck_channels
        self._total_key_depth = total_key_depth
        self._total_value_depth = total_value_depth
        self._num_heads = num_heads
        self._bn_layer = bn_layer
        self._conv_kernel_weight_decay = conv_kernel_weight_decay
        self._activation = activation
        self._activation_fn = activations.get_activation(activation)
        self._feed_forward_network_channels = feed_forward_network_channels

    def build(self, input_shape_list):
        pixel_shape, memory_shape = input_shape_list[:2]
        # Here we follow ResNet bottleneck blocks: we apply a batch norm with gamma
        # initialized at zero, followed by drop path and an activation function.
        # Initializing this gamma at zero ensures that at random initialization of
        # the model, the skip connections dominate all residual blocks. In this way,
        # all the skip connections construct an identity mapping that passes the
        # gradients (without any distortion from the randomly initialized blocks) to
        # all residual blocks. This helps training at early epochs.
        # Reference: "Accurate, Large Minibatch SGD: Training ImageNet in 1 Hour".
        # https://arxiv.org/abs/1706.02677
        if (self._use_memory2pixel_feedback_attention or
                self._use_memory_self_attention):
            self._memory_conv3_bn = convolutions.Conv1D(
                memory_shape[-1], 'memory_conv3_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=self._bn_layer,
                bn_gamma_initializer='zeros',
                activation='none',
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

        if self._feed_forward_network_channels > 0:
            self._memory_ffn_conv1_bn_act = convolutions.Conv1D(
                self._feed_forward_network_channels, 'memory_ffn_conv1_bn_act',
                use_bias=False,
                use_bn=True,
                bn_layer=self._bn_layer,
                activation=self._activation,
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)
            # Again, we follow ResNet bottleneck blocks: we apply a batch norm with
            # gamma initialized at zero, followed by drop path and an activation
            # function.
            self._memory_ffn_conv2_bn = convolutions.Conv1D(
                memory_shape[-1], 'memory_ffn_conv2_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=self._bn_layer,
                bn_gamma_initializer='zeros',
                activation='none',
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)
        if self._use_pixel2memory_feedback_attention:
            self._pixel_conv3_bn = convolutions.Conv1D(
                pixel_shape[-1], 'pixel_conv3_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=self._bn_layer,
                bn_gamma_initializer='zeros',
                activation='none',
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

        if self._use_kmeans_cross_attention:
            self._kmeans_memory_batch_norm_retrieved_value = self._bn_layer(
                axis=-1, name='kmeans_memory_batch_norm_retrieved_value')
            self._kmeans_memory_conv3_bn = convolutions.Conv1D(
                memory_shape[-1], 'kmeans_memory_conv3_bn',
                use_bias=False,
                use_bn=True,
                bn_layer=self._bn_layer,
                bn_gamma_initializer='zeros',
                activation='none',
                conv_kernel_weight_decay=self._conv_kernel_weight_decay)

    def kmeans_assignment_step(self, pixel_feature, cluster_centers,
                               num_mask_slots, training):
        auxiliary_result_dict = self._auxiliary_clustering_predictor(
            {
                'feature_panoptic': pixel_feature,
                'feature_semantic': pixel_feature,
                'transformer_class_feature': cluster_centers,
                'transformer_mask_feature': cluster_centers
            },
            training=training)
        # A cluster-wise argmax is applied to convert the attention logits to
        # clustering results, which serve as attention weights.
        clustering_result = tf.argmax(
            auxiliary_result_dict[common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY], axis=-1)
        clustering_result = tf.cast(
            tf.one_hot(clustering_result, depth=num_mask_slots, axis=-1),
            tf.float32)
        clustering_result = tf.stop_gradient(clustering_result)
        return auxiliary_result_dict, clustering_result

    def kmeans_update_memory_step(self, clustering_result, pixel_value, training):
        kmeans_memory_space = tf.einsum('blm,bld->bmd', clustering_result,
                                        pixel_value)
        kmeans_memory_space = self._kmeans_memory_batch_norm_retrieved_value(
            kmeans_memory_space, training=training)
        kmeans_memory_space = self._kmeans_memory_conv3_bn(
            kmeans_memory_space, training=training)
        return kmeans_memory_space

    def call(self, inputs):
        """Performs a forward pass.

        We have to define drop_path_masks outside the layer call and pass it into
        the layer call, because recompute_grad (gradient checkpointing) does not
        allow any randomness within the function call. In addition, recompute_grad
        only supports float tensors as inputs. For this reason, the training flag
        should be also passed as a float tensor. For the same reason, we cannot
        support passing drop_path_random_mask as None. Instead, we ask the users to
        pass only the first two tensors when drop path is not used.

        Args:
          inputs: A tuple of 4 or 8 tensors, containing
            pixel_space_input should be a [batch, height, width,
              pixel_space_channels] tensor.
            memory_space_input should be a [batch, num_memory,
              memory_space_channels] tensor.
            auxiliary_outputs should be a tuple containing auxiliary outputs, where
              each element has the dictionary type.
            float_tensor_training should be a float tensor of 0.0 or 1.0, whether
              the model is in training mode.
            (optional) pixel_space_drop_path_mask is a drop path mask tensor of
              shape [batch, 1, 1] for the pixel space.
            (optional) memory_space_attention_drop_path_mask is a drop path mask
              tensor of shape [batch, 1, 1] for the memory space.
            (optional) memory_kmeans_attention_drop_path_mask is a drop path mask
              tensor of shape [batch, 1, 1] for the memory space.
            (optional) memory_space_feed_forward_network_drop_path_mask is a drop
              path mask tensor of shape [batch, 1, 1] for the memory space feed
              forward network.

        Returns:
          pixel_space_output: A [batch, num_pixel, pixel_space_channels] tensor.
          memory_space_output: A [batch, num_memory, memory_space_channels]
            tensor.
          auxiliary_outputs: A tuple containing auxiliary outputs, where each
            element has the dictionary type.

        Raises:
          ValueError: If the length of inputs is not 4 or 8.
        """
        if len(inputs) not in (4, 8):
            raise ValueError('The length of inputs should be either 4 or 8.')

        # Unpack the inputs.
        (pixel_space_input, memory_space_input, auxiliary_outputs,
         float_tensor_training,
         pixel_space_drop_path_mask, memory_space_attention_drop_path_mask,
         memory_kmeans_attention_drop_path_mask,
         memory_space_feed_forward_network_drop_path_mask) = (
            utils.pad_sequence_with_none(inputs, target_length=8))

        # Recompute_grad takes only float tensors as inputs. It does not allow
        # bools or boolean tensors. For this reason, we cast training to a float
        # tensor outside this call, and now we cast it back to a boolean tensor.
        training = tf.cast(float_tensor_training, tf.bool)

        # Decode the inputs shapes.
        pixel_shape = pixel_space_input.get_shape().as_list()
        memory_shape = memory_space_input.get_shape().as_list()

        # Flatten the pixel_space_input.
        pixel_space_input = tf.reshape(
            pixel_space_input,
            [-1, pixel_shape[1] * pixel_shape[2], pixel_shape[3]])

        # Similar to the ResNet bottleneck design, we do an input down projection
        # in both the pixel space and the memory space.
        memory_space = self._memory_conv1_bn_act(memory_space_input,
                                                 training=training)

        # Pixel space input is not activated.
        pixel_space = self._pixel_conv1_bn_act(
            self._activation_fn(pixel_space_input), training=training)

        if (self._use_memory_self_attention or
                self._use_pixel2memory_feedback_attention):
            memory_space_qkv = self._memory_qkv_conv_bn(memory_space,
                                                        training=training)
            # Split, reshape, and transpose the query, key, and value.
            memory_query, memory_key, memory_value = (
                tf.split(memory_space_qkv, [
                    self._total_key_depth, self._total_key_depth,
                    self._total_value_depth], axis=-1))
            memory_key = utils.reshape_and_transpose_for_attention_operation(
                memory_key, self._num_heads)
            memory_value = tf.reshape(memory_value, [
                -1, memory_shape[1], self._num_heads,
                self._total_value_depth // self._num_heads])
            # Reshape and transpose the query.
            memory_query = utils.reshape_and_transpose_for_attention_operation(
                memory_query, self._num_heads)
        elif self._use_memory2pixel_feedback_attention:
            # Compute memory query only if memory key and value are not used.
            memory_query = self._memory_query_conv_bn(memory_space, training=training)
            # Reshape and transpose the query.
            memory_query = utils.reshape_and_transpose_for_attention_operation(
                memory_query, self._num_heads)

        if self._use_pixel2memory_feedback_attention:
            pixel_space_qkv = self._pixel_qkv_conv_bn(pixel_space,
                                                      training=training)
            # Split the query, key, and value.
            pixel_query, pixel_key, pixel_value = tf.split(
                pixel_space_qkv, [
                    self._total_key_depth, self._total_key_depth,
                    self._total_value_depth], axis=-1)
            pixel_query = utils.reshape_and_transpose_for_attention_operation(
                pixel_query, self._num_heads)
            # Reshape and transpose the key and the value.
            pixel_key = utils.reshape_and_transpose_for_attention_operation(
                pixel_key, self._num_heads)
        elif self._use_memory2pixel_feedback_attention:
            pixel_space_kv = self._pixel_kv_conv_bn(pixel_space, training=training)
            # Split the key and the value.
            pixel_key, pixel_value = tf.split(pixel_space_kv, [
                self._total_key_depth, self._total_value_depth], axis=-1)
            # Reshape and transpose the key and the value.
            pixel_key = utils.reshape_and_transpose_for_attention_operation(
                pixel_key, self._num_heads)
        else:
            # In this case, only k-means cross attention is enabled, and thus we only
            # need the value of pixel to update memory.
            pixel_value = self._pixel_v_conv_bn(pixel_space, training=training)
        pixel_value = tf.reshape(pixel_value, [
            -1, pixel_shape[1] * pixel_shape[2], self._num_heads,
            self._total_value_depth // self._num_heads])

        memory_space_output = memory_space_input
        pixel_space_output = pixel_space_input

        # Perform kmeans cross-attention.
        if self._use_kmeans_cross_attention:
            # Assignment step.
            pixel_space2d = tf.reshape(
                pixel_space,
                [-1, pixel_shape[1], pixel_shape[2], self._bottleneck_channels])
            auxiliary_result_dict, clustering_result = self.kmeans_assignment_step(
                pixel_space2d, memory_space, memory_shape[1], training)
            clustering_result = tf.reshape(clustering_result,
                                           [-1, pixel_shape[1] * pixel_shape[2],
                                            memory_shape[1]])
            # Add to auxiliary_outputs.
            auxiliary_outputs = auxiliary_outputs + (auxiliary_result_dict,)
            # Update step.
            pixel_value_single_head = tf.reshape(
                pixel_value,
                [-1, pixel_shape[1] * pixel_shape[2], self._total_value_depth])
            kmeans_memory_space = self.kmeans_update_memory_step(
                clustering_result, pixel_value_single_head, training)
            if memory_kmeans_attention_drop_path_mask is not None:
                kmeans_memory_space = (
                        kmeans_memory_space * memory_kmeans_attention_drop_path_mask)
            memory_space_output = memory_space_output + kmeans_memory_space

        # Perform M2P/M2M attention.
        # Compute memory space attention.
        memory_attention_key = []
        memory_attention_value = []
        if self._use_memory2pixel_feedback_attention:
            # If memory self attention is not used, then only memory2pixel cross
            # attention is used for the memory space. In this case, the key and the
            # value are simply pixel_key and pixel_value.
            memory_attention_key.append(pixel_key)
            memory_attention_value.append(pixel_value)
        if self._use_memory_self_attention:
            memory_attention_key.append(memory_key)
            memory_attention_value.append(memory_value)

        if (self._use_memory2pixel_feedback_attention or
                self._use_memory_self_attention):
            # If we also use memory self attention, the key and the value are the
            # concatenation of keys and values in both the pixel space and the
            # memory space.
            memory_attention_key = tf.concat(memory_attention_key, axis=2)
            memory_attention_value = tf.concat(memory_attention_value, axis=1)
            memory_space = self._memory_attention(
                (memory_query, memory_attention_key, memory_attention_value),
                training=training)
            memory_space = self._memory_conv3_bn(memory_space, training=training)

            if memory_space_attention_drop_path_mask is not None:
                memory_space = memory_space * memory_space_attention_drop_path_mask
            memory_space_output = memory_space_output + memory_space

        memory_space_output = self._activation_fn(memory_space_output)

        # Apply an optional feed-forward network to the memory space.
        if self._feed_forward_network_channels > 0:
            memory_space = self._memory_ffn_conv1_bn_act(memory_space_output,
                                                         training=training)
            memory_space = self._memory_ffn_conv2_bn(memory_space,
                                                     training=training)
            if memory_space_feed_forward_network_drop_path_mask is not None:
                memory_space = (memory_space *
                                memory_space_feed_forward_network_drop_path_mask)
            memory_space_output = self._activation_fn(
                memory_space_output + memory_space)

        # Perform P2M attention.
        # Compute pixel space attention and the output projection only when
        # pixel2memory_feedback_attention is used.
        if self._use_pixel2memory_feedback_attention:
            pixel_space = self._pixel_attention(
                (pixel_query, memory_key, memory_value), training=training)
            pixel_space = self._pixel_conv3_bn(pixel_space, training=training)
            if pixel_space_drop_path_mask is not None:
                pixel_space = pixel_space * pixel_space_drop_path_mask
            pixel_space_output = pixel_space_input + pixel_space

        # Reshape back to 4D.
        pixel_space_output = tf.reshape(
            pixel_space_output,
            [-1, pixel_shape[1], pixel_shape[2], pixel_shape[3]])

        # Return the pixel space output, memory space output, and auxiliary outputs.
        return pixel_space_output, memory_space_output, auxiliary_outputs



"""
Note: even though this is called old-polygon-extraction, it was never
fully refactored out of the critical path. Due to funding cuts we arent
investing effort in changing this. A refactor is needed, but for now that will
just have to add to our our technical dept.

The original heatmap -> polygon extraction code.
"""
import ubelt as ub
import itertools
import scriptconfig as scfg

from geowatch.heuristics import SITE_SUMMARY_CNAME
from geowatch.tasks.tracking import agg_functions

from geowatch.tasks.tracking.utils import (
    mask_to_polygons,
    score_track_polys,
    gpd_len)


class FoundNothing(Exception):
    """
    Error to help short-circuit the part where we return an empty list
    """
    ...


class PolygonExtractConfig(scfg.DataConfig):
    # This is the base config that all from-heatmap trackers have in common
    # which has to do with how heatmaps are loaded, normalized, and aggregated.
    # This is associated with :func:`_gids_polys`

    new_algo = scfg.Value(None, help=ub.paragraph(
        '''
        If None, use the old algorithm, otherwise use one of the new algorithm.
        This is a hack that needs to get redesigned and integrated in a better
        way.
        '''))

    key = scfg.Value('salient', help=ub.paragraph(
        '''
        One or more channels to use as positive class for binary heatmap
        polygon extraction and scoring.
        '''))

    agg_fn = scfg.Value('probs', help=ub.paragraph(
        '''
        The aggregation method to preprocess heatmaps.
        See ``agg_functions.AGG_FN_REGISTRY`` for available options.
        (3d heatmaps -> 2d heatmaps), calling convention TBD
        '''), alias=['outer_agg_fn'])

    thresh = scfg.Value(0.0, help=ub.paragraph(
        '''
        The threshold for polygon extraction from heatmaps.
        E.g. this threshold binarizes the heatmaps.

        For each frame, if sum of foreground heatmaps > thresh,
            class is max(foreground keys).
            else, class is max(background keys).
        '''))

    morph_kernel = scfg.Value(3, help=ub.paragraph(
        '''
        Morphology kernel for preprocessing the heatmaps with dilation.
        Height/width in px of close or dilate kernel
        '''))

    thresh_hysteresis = scfg.Value(None, help=ub.paragraph(
        '''
        I dont remember. Help wanted to document this, or remove it.
        '''))

    # TODO: Consolidate into agg_fn
    norm_ord = scfg.Value(1, help=ub.paragraph(
        '''
        The generalized mean order used to average heatmaps over the
        "outer_window_size". A value of 1 is the normal mean. A value of inf
        is the max function. Note: this is effectively an outer_agg_fn.

        order of norm to aggregate heatmap pixels across time.
        1: average [default]
        2: euclidean
        0: sum
        np.inf, 'inf', or None: max
        '''))

    # TODO: rename to outer_window_size
    moving_window_size = scfg.Value(None, help=ub.paragraph(
        '''
        The outer moving window size. The number of consecutive inner window
        results to aggregate together. If None, then all inner window results
        are combined into a single final heatmap.
        '''), alias=['outer_window_size'])

    inner_window_size = scfg.Value(None, help=ub.paragraph(
        '''
        The inner moving window time range (e.g. 1y).  The bucket size (in
        time) of time-consecutive heatmaps to combine using an inner moving
        window. If None, then no inner windowing is used.
        '''))

    inner_agg_fn = scfg.Value('mean', help=ub.paragraph(
        '''
        The method used for aggregating heatmaps scores over the inner window.
        Note, this roughtly corresponds to norm_ord, which is like the
        outer_agg_fn.
        '''))

    resolution = scfg.Value(None, help=ub.paragraph(
        '''
        The resolution for loading and processing the heatmaps at. E.g. 10GSD.
        '''))

    dynamic_fixed_resolution = scfg.Value(None, type=str, help=ub.paragraph(
        '''
        Experimental. Added in 0.17.1
        Similar to the variant in kwcoco video dataset.
        example value: {'max_trkspace_full_dims': [1000, 1000]}
        '''))

    use_boundaries = scfg.Value(False, help=ub.paragraph(
        '''
        If False, then extracted polygons are used as new site boundaries.  If
        True, then we keep existing annotation boundaries unchanged and only
        used the heatmaps to update scores of the existing boundary polyons.
        '''))

    poly_merge_method = scfg.Value('v1', help=ub.paragraph(
        '''
        Method for handling overlaping polygons across multiple timesteps.
        Currently can be "v1" or "v2". There isn't much difference.
        We should find a better way of handling this.
        '''))

    polygon_simplify_tolerance = scfg.Value(None, help=ub.paragraph(
        '''
        The pixel size (at the specified heatmap resolution) to use for polygon
        simplification.
        '''))

    viz_out_dir = scfg.Value(None, help=ub.paragraph(
        '''
        Directory to output intermediate visualizations
        '''))


def _gids_polys(sub_dset, video_id, **kwargs):
    """
    This is associated with :class:`PolygonExtractConfig`

    Example:
        >>> from geowatch.tasks.tracking.old_polygon_extraction import *  # NOQA
        >>> from geowatch.tasks.tracking.old_polygon_extraction import _gids_polys
        >>> import geowatch
        >>> coco_dset = geowatch.coerce_kwcoco(data='geowatch-msi', dates=True, geodata=True, heatmap=True)
        >>> sub_dset = coco_dset.subset(coco_dset.videos().images[0])
        >>> video_id = list(sub_dset.videos())[0]
        >>> kwargs = dict(key=['salient'],
        >>>     agg_fn='probs',
        >>>     thresh=0.001,
        >>>     morph_kernel=3,
        >>>     thresh_hysteresis=0,
        >>>     norm_ord=1,
        >>>     resolution=None,
        >>>     outer_window_size=None,
        >>>     use_boundaries=None)
        >>> results1 = list(_gids_polys(sub_dset, video_id, **kwargs))
        >>> kwargs['new_algo'] = 'crall'
        >>> results2 = list(_gids_polys(sub_dset, video_id, **kwargs))
        >>> print(f'results1 = {ub.urepr(results1, nl=1)}')
        >>> print(f'results2 = {ub.urepr(results2, nl=1)}')

    Example:
        >>> from geowatch.tasks.tracking.old_polygon_extraction import *  # NOQA
        >>> from geowatch.tasks.tracking.old_polygon_extraction import _gids_polys
        >>> import geowatch
        >>> coco_dset = geowatch.coerce_kwcoco(data='geowatch-msi', dates=True, geodata=True, heatmap=True)
        >>> sub_dset = coco_dset.subset(coco_dset.videos().images[0])
        >>> video_id = list(sub_dset.videos())[0]
        >>> kwargs = dict(key=['salient'],
        >>>     agg_fn='probs',
        >>>     thresh=0.001,
        >>>     morph_kernel=3,
        >>>     thresh_hysteresis=0,
        >>>     norm_ord=1,
        >>>     resolution=None,
        >>>     dynamic_fixed_resolution={'max_trkspace_full_dims': [16, 16]},
        >>>     outer_window_size=None,
        >>>     use_boundaries=None)
        >>> results1 = list(_gids_polys(sub_dset, video_id, **kwargs))
        >>> kwargs['new_algo'] = 'crall'
        >>> results2 = list(_gids_polys(sub_dset, video_id, **kwargs))
        >>> print(f'results1 = {ub.urepr(results1, nl=1)}')
        >>> print(f'results2 = {ub.urepr(results2, nl=1)}')

    Returns:
        Iterable[Tuple[List[int], MultiPolygon]] -
            For each track return a list of image ids and a single associated
            polygon.
    """
    from kwutil import util_time
    import numpy as np
    import rich
    config = PolygonExtractConfig(**kwargs)

    requested_resolution = config.resolution

    import kwutil
    dynamic_fixed_resolution = kwutil.util_yaml.Yaml.coerce(config.dynamic_fixed_resolution)
    """
    dynamic_fixed_resolution = {'max_trkspace_full_dims': [1, 1]}
    """
    if dynamic_fixed_resolution is not None:
        # _debug = 1
        # if _debug:
        #     print(f'dynamic_fixed_resolution = {ub.urepr(dynamic_fixed_resolution, nl=1)}')
        if isinstance(dynamic_fixed_resolution, dict):
            from geowatch.utils.util_resolution import ResolvedWindow
            from geowatch.utils.util_resolution import ResolvedUnit
            video = sub_dset.index.videos[video_id]
            vid_w = video['width']
            vid_h = video['height']
            target_gsd = video['target_gsd']
            vidspace_full_dims = ResolvedWindow.coerce(f'{vid_w} x {vid_h} @ {target_gsd} GSD')
            if requested_resolution is None:
                trkspace_full_dims = vidspace_full_dims
                trk_gsd = vidspace_full_dims.resolution
            else:
                trk_gsd = ResolvedUnit.coerce(requested_resolution, default_unit='GSD')
                trkspace_full_dims = vidspace_full_dims.at_resolution(trk_gsd)

            max_trkspace_full_dims = dynamic_fixed_resolution.get('max_trkspace_full_dims', None)
            if max_trkspace_full_dims is not None:
                max_trkspace_full_dims = np.array(max_trkspace_full_dims)
                trkspace_full_dims = np.array(trkspace_full_dims.window)
                dynamic_factor = (max_trkspace_full_dims / trkspace_full_dims).min()
                _debug = 1
                if dynamic_factor < 1.0:
                    if _debug:
                        print('---')
                        print('Handle dynamic resolution adjustment')
                        print(f'max_trkspace_full_dims = {ub.urepr(max_trkspace_full_dims, nl=1)}')
                        print(f'before: trkspace_full_dims={trkspace_full_dims}')
                        print(f'before: trk_gsd = {ub.urepr(trk_gsd, nl=1)}')
                    # Adjust the scale
                    trk_gsd = (trk_gsd / dynamic_factor)
                    if _debug:
                        print(f'dynamic_factor = {ub.urepr(dynamic_factor, nl=1)}')
                        print(f'after: trk_gsd = {ub.urepr(trk_gsd, nl=1)}')
                        print('---')
                else:
                    dynamic_factor = 1.0
                trk_resolution = trk_gsd.__nice__()
                # raise NotImplementedError
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError('todo')
    else:
        trk_resolution = requested_resolution

    print(f'config.use_boundaries={config.use_boundaries}')
    if config.use_boundaries:  # for AC/SC
        raw_boundary_tracks = score_track_polys(sub_dset, video_id, [SITE_SUMMARY_CNAME],
                                                resolution=trk_resolution)

        if len(raw_boundary_tracks) == 0:
            gids = sub_dset.images(video_id=video_id).gids

            print(f'SITE_SUMMARY_CNAME={SITE_SUMMARY_CNAME}')
            print(f'trk_resolution={trk_resolution}')
            print(f'sub_dset={sub_dset}')
            print(f'video_id={video_id}')
            # anns = sub_dset.annots(video_id=video_id)
            anns = sub_dset.annots()
            set(anns.images.lookup('video_id'))
            boundary_tracks = [(None, None)]
            import warnings
            msg = ('need valid site boundaries!')
            warnings.warn(msg)
            # raise AssertionError(msg)
            raise FoundNothing(msg)
        else:
            gids = raw_boundary_tracks['gid'].unique()
            print('generating polys in bounds: number of bounds: ',
                  gpd_len(raw_boundary_tracks))
            boundary_tracks = list(raw_boundary_tracks.groupby('track_idx'))

    else:
        boundary_tracks = [(None, None)]
        # TODO WARNING this is wrong!!! need to make sure this is never used.
        # The gids are lexically sorted, not sorted by order in video!
        # gids = list(sub_dset.imgs.keys())
        # vidid = list(sub_dset.index.vidid_to_gids.keys())[0]
        gids = sub_dset.images(video_id=video_id).gids

    images = sub_dset.images(gids)
    image_dates = [util_time.coerce_datetime(d)
                   for d in images.lookup('date_captured', None)]
    # image_years = [d.year for d in image_dates]

    channels_list = config.key
    channels = '|'.join(config.key)
    coco_images = images.coco_images

    load_workers = 0  # TODO: configure
    load_jobs = ub.JobPool(mode='process', max_workers=load_workers)

    print(f'Reading heatmaps with: channels={channels}')

    with load_jobs:

        delayed_images = []
        for coco_img in coco_images:
            delayed = coco_img.imdelay(channels=channels, space='video', resolution=trk_resolution)
            delayed_images.append(delayed)

        # Some configurations can blow out memory here.
        # Check how much data would be loaded before we do it.
        CHECK_LOAD_MEMORY = 1
        if CHECK_LOAD_MEMORY:
            from geowatch.utils import util_units
            from geowatch.utils import util_hardware
            mem_info = util_hardware.get_mem_info(with_units=True)
            bytes_per_cell = np.dtype('float32').itemsize
            num_raster_cells = sum([np.prod(delayed.shape) for delayed in delayed_images])
            total_bytes = num_raster_cells * bytes_per_cell
            ureg = util_units.unit_registry()
            total_size = (total_bytes * ureg.bytes).to('gigabytes')
            print(f'Loading heatmaps will use {total_size} in memory')
            print(f'mem_info = {ub.urepr(mem_info, nl=1)}')

            # TODO: dynamic resolution reduction
            if total_size.m > 32:
                print('WARNING: heatmaps will take more than 32GB of memory. '
                      'You may want to lower resolution')

            problem_likelihood = ''
            if mem_info['available'] < total_size:
                problem_likelihood = 'is VERY likely to overload memory'
            elif mem_info['available'] < (total_size * 1.5):
                problem_likelihood = 'is likely to overload memory'
            elif mem_info['available'] < (total_size * 2):
                # Warn if we don't have twice as much space, this accounts for
                # things like copies, which may happen.
                problem_likelihood = 'uncomfortably close to overloading memory'
            elif mem_info['available'] < (total_size * 2.5):
                # Warn if we don't have twice as much space, this accounts for
                # things like copies, which may happen.
                problem_likelihood = 'may overload memory'
            if problem_likelihood:
                print(f'mem_info = {ub.urepr(mem_info, nl=1)}')
                msg = ub.paragraph(
                    '''
                    !!! Warning !!!
                    The set of images about to be loaded {problem_likelihood}
                    ''')
                print(msg)
                warnings.warn(msg)

        for delayed in delayed_images:
            load_jobs.submit(delayed.finalize)

        _heatmaps = []
        for job in ub.ProgIter(load_jobs.jobs, desc='Load heatmap'):
            _heatmap = job.result()
            _heatmaps.append(_heatmap)

    _heatmaps_thwc = np.stack(_heatmaps, axis=0)

    if config.new_algo is not None:
        # HACK in new algo
        import kwimage
        if config.use_boundaries:
            # HACK TO LOAD BOUNDS FOR POLYGONS
            from geowatch.tasks.tracking.utils import _build_annot_gdf
            cnames = [SITE_SUMMARY_CNAME]
            aids = list(ub.flatten(images.annots))
            gdf, flat_scales = _build_annot_gdf(sub_dset, aids=aids, cnames=cnames, resolution=trk_resolution)
            if len(gdf) == 0:
                # assert len(gdf) > 0, 'need valid site boundaries!'
                bounds = None
            else:
                union_poly = gdf.unary_union
                bounds = kwimage.MultiPolygon.from_shapely(union_poly)
        else:
            bounds = None

        video_name = sub_dset.index.videos[video_id]['name']
        if config.viz_out_dir is None:
            extractor_viz_dir = None
        else:
            extractor_viz_dir = ub.Path(config.viz_out_dir) / video_name

        from geowatch.tasks.tracking import polygon_extraction
        import kwcoco
        classes = kwcoco.CategoryTree.from_mutex(channels_list)
        extractor = polygon_extraction.PolygonExtractor(
            _heatmaps_thwc,
            heatmap_time_intervals=image_dates,
            bounds=bounds,
            classes=classes,
            config={
                'scale_factor': 1,
                'thresh': config.thresh,
                'algo': config.new_algo,
                'viz_out_dir': extractor_viz_dir,
                'polygon_simplify_tolerance': config.polygon_simplify_tolerance,
            })
        polygons = extractor.predict_polygons()

        # Conform to expected output
        result_gen = []
        for poly in polygons:
            single_result = (gids, poly)
            result_gen.append(single_result)
    else:
        print(f'(presum) _heatmaps_thwc.shape={_heatmaps_thwc.shape}')
        _heatmaps = _heatmaps_thwc.sum(axis=-1)  # sum over channels
        print(f'_heatmaps.shape={_heatmaps.shape}')
        missing_ix = np.array([channels not in i.channels for i in coco_images])

        num_missing = missing_ix.sum()
        num_images  = len(missing_ix)
        if num_missing > 0:
            rich.print(f'[yellow]There are {num_missing} / {num_images} images that are missing {channels} channels')
        else:
            rich.print(f'[green]There are {num_missing} / {num_images} images that are missing {channels} channels')

        # TODO this was actually broken in orig, so turning it off here for now
        interpolate = 0
        if interpolate:
            diffed = np.concatenate((np.diff(missing_ix), [False]))
            src = ~missing_ix & diffed
            _heatmaps[missing_ix] = _heatmaps[src]
            if missing_ix[0]:
                _heatmaps[:np.searchsorted(diffed, True)] = 0
            assert np.isnan(_heatmaps).all(axis=(1, 2)).sum() == 0
        else:
            _heatmaps[missing_ix] = 0

        # no benefit so far
        proc_jobs = ub.JobPool('process', max_workers=0)
        with proc_jobs:

            for _, track in ub.ProgIter(boundary_tracks, desc='submit proc jobs'):
                proc_jobs.submit(_process, track, _heatmaps, image_dates, gids, config)

            result_gen = itertools.chain.from_iterable(
                j.result() for j in ub.ProgIter(proc_jobs.jobs, desc='collect proc jobs'))
            result_gen = list(result_gen)

    # If we did dynamic resolution, unscale the polygons.
    if dynamic_fixed_resolution is not None:
        # Hacky, might want to handle the dynamic resolution at the layer above
        # this.
        if dynamic_factor < 1:
            undo_factor = 1 / dynamic_factor
            result_gen2 = []
            for gids, poly in result_gen:
                poly = poly.scale(undo_factor)
                result_gen2.append((gids, poly))
            result_gen = result_gen2

    return result_gen


def _process(track, _heatmaps, image_dates, gids, config):
    """
    Yields:
        Tuple[List[int], MultiPolygon] -
            a list of image ids a polygon is valid for, and
            the single polygon corresponding.
    """
    from shapely.ops import unary_union
    import kwimage
    import numpy as np
    # TODO when bounds are time-varying, this lets individual frames
    # go outside them; only enforces the union. Problem?
    # currently bounds come from site summaries, which are not
    # time-varying.
    if track is None:
        track_bounds = None
        _heatmaps_in_track = _heatmaps
        heatmap_dates = image_dates
    else:
        track_bounds = track['poly'].unary_union
        track_gids = track['gid']
        flags = np.in1d(gids, track_gids)
        _heatmaps_in_track = np.compress(flags, _heatmaps, axis=0)
        heatmap_dates = list(ub.compress(image_dates, flags))

    # this is another hot spot, heatmaps_to_polys -> mask_to_polygons ->
    # rasterize. Figure out how to vectorize over bounds.
    track_polys = heatmaps_to_polys(
        _heatmaps_in_track, track_bounds, heatmap_dates=heatmap_dates,
        config=config,
    )
    if track is None:
        # BUG: The polygons retunred from heatmap-to-polys might not be
        # corresponding to the gids in this case.
        yield from zip(itertools.repeat(gids), track_polys)
        # for poly in polys:  # convert to shapely to check this
        # if poly.is_valid and not poly.is_empty:
        # yield (gids, poly)
    else:
        poly = unary_union([p.to_shapely() for p in track_polys])
        if poly.is_valid and not poly.is_empty:
            yield (track['gid'], kwimage.MultiPolygon.from_shapely(poly))


def heatmaps_to_polys(heatmaps, track_bounds, heatmap_dates=None, config=None):
    """
    Use parameters: agg_fn, thresh, morph_kernel, thresh_hysteresis, norm_ord

    Args:
        heatmaps (ndarray): A [T, H, W] heatmap

        track_bounds (kwimage.MultiPolygon | None):
            a valid region in the heatmaps where new polygons can be extracted.

        heatmap_dates (List[datetime] | None):
            dates corresponding with each heatmap time dimension

        config (PolygonExtractConfig): polygon extraction config

    Example:
        >>> from geowatch.tasks.tracking.old_polygon_extraction import *  # NOQA
        >>> import kwimage
        >>> from kwutil import util_time
        >>> import numpy as np
        >>> from geowatch.tasks.tracking.old_polygon_extraction import PolygonExtractConfig  # NOQA
        >>> config = PolygonExtractConfig()
        >>> heatmaps = np.zeros((7, 64, 64))
        >>> heatmaps[2, 20:40, 20:40] = 1
        >>> heatmaps[5, 30:50, 30:50] = 1
        >>> heatmap_dates = [util_time.coerce_datetime(x) for x in [
        >>>     '2020-01-01', '2020-02-01', '2020-03-01', '2020-04-01',
        >>>     '2020-05-01', '2020-06-01', '2020-07-01', ]]
        >>> track_bounds = kwimage.Polygon.random(rng=0).scale((64, 64))
        >>> # V1 merges everything together across all time
        >>> config.poly_merge_method = 'v1'
        >>> polygons_final = heatmaps_to_polys(heatmaps, track_bounds, heatmap_dates=heatmap_dates, config=config)
        >>> # V3 does some time separation
        >>> config.poly_merge_method = 'v3'
        >>> polygons_final = heatmaps_to_polys(heatmaps, track_bounds, heatmap_dates=heatmap_dates, config=config)
    """
    from kwutil import util_time
    import numpy as np

    # TODO: rename moving window size to "outer_window_size"

    def convert_to_shapely(polys):
        return [p.to_shapely() for p in polys]

    def convert_to_kwimage_poly(shapely_polys):
        import kwimage
        return [kwimage.Polygon.from_shapely(p) for p in shapely_polys]

    # outer agg function
    _agg_fn = agg_functions.AGG_FN_REGISTRY[config.agg_fn]

    if config.inner_window_size is not None:
        # TODO: generalize if needed
        assert heatmap_dates is not None

        if config.inner_agg_fn == 'mean':
            inner_ord = 1
        elif config.inner_agg_fn == 'max':
            inner_ord = float('inf')
        else:
            raise NotImplementedError(config.inner_agg_fn)

        # Do inner aggregation before outer aggregation
        groupxs = _compute_time_window(
            config.inner_window_size, num_frames=len(heatmaps),
            heatmap_dates=heatmap_dates)

        new_heatmaps = []
        new_intervals = []
        for idxs in groupxs:
            new_start_date = min(ub.take(heatmap_dates, idxs))
            new_end_date = max(ub.take(heatmap_dates, idxs))
            inner = agg_functions._norm(heatmaps[idxs], norm_ord=inner_ord)
            new_intervals.append((new_start_date, new_end_date))
            new_heatmaps.append(inner)

        heatmap_date_intervals = new_intervals
        heatmaps = np.array(new_heatmaps)
    else:
        if config.inner_window_size is not None:
            raise NotImplementedError(
                'only temporal deltas for inner agg window for now')

        if heatmap_dates is None:
            heatmap_dates = [util_time.coerce_datetime(0)] * len(heatmaps)
        heatmap_date_intervals = [(t, t) for t in heatmap_dates]

    heatmap_unixtime_intervals = np.array([
        (a.timestamp(), b.timestamp())
        for a, b in heatmap_date_intervals
    ], dtype=np.float64)
    heatmap_dates = [util_time.coerce_datetime(a) for (a, b) in
                     heatmap_unixtime_intervals]

    # calculate number of moving-window steps, based on window_size and number
    # of heatmaps
    if config.moving_window_size is None:
        # Hack: none meant window size is infinite here.
        groupxs = [np.arange(len(heatmaps))]
    else:
        groupxs = _compute_time_window(
            config.moving_window_size, num_frames=len(heatmaps),
            heatmap_dates=heatmap_dates)

    # initialize heatmaps and initial polygons on the first set of heatmaps
    n_steps = len(groupxs)
    xs_init = groupxs[0]
    h_init = heatmaps[xs_init]
    t_init = heatmap_unixtime_intervals[xs_init]

    # print(f'n_steps={n_steps}')
    # print('!!!!')
    prog = ub.ProgIter(total=n_steps, desc='process-step')
    # prog.begin()
    with prog:
        step_idx = 0
        polys_final = _process_1_step(h_init, _agg_fn, track_bounds, step_idx, config)
        times_final = [[t_init[0][0], t_init[-1][1]]] * len(polys_final)
        prog.step()

        if n_steps > 1:
            polys_final = convert_to_shapely(polys_final)

            for step_idx in range(1, n_steps):
                idxs = groupxs[step_idx]
                prog.step()
                prog.ensure_newline()
                h1 = heatmaps[idxs]
                t1 = heatmap_unixtime_intervals[idxs]
                print(h1.sum())

                p1 = _process_1_step(h1, _agg_fn, track_bounds, step_idx, config)
                t1 = [[t1[0][0], t1[-1][1]]] * len(p1)
                p1 = convert_to_shapely(p1)

                polys_final, times_final = _merge_polys(
                    polys_final, times_final,
                    p1, t1,
                    poly_merge_method=config.poly_merge_method,
                )

            polys_final = convert_to_kwimage_poly(polys_final)
    return polys_final


def _compute_time_window(window, num_frames=None, heatmap_dates=None):
    """
    Groups frame indexes into "inner" or "outer" windows.

    Example:
        >>> window = 5
        >>> num_frames = 23
        >>> groupxs = _compute_time_window(window, num_frames)
        >>> print(f'groupxs={groupxs}')
        >>> #
        >>> window = '7days'
        >>> from kwutil import util_time
        >>> heatmap_dates = list(map(util_time.coerce_datetime, [
        >>>     '2020-01-01', '2020-01-02', '2020-02-01',
        >>>     '2020-02-02', '2020-03-14', '2020-03-23',
        >>>     '2020-04-01', '2020-06-23', '2020-06-26',
        >>>     '2020-06-27', '2020-06-28', ]))
        >>> groupxs = _compute_time_window(window, num_frames, heatmap_dates)
        >>> print(f'groupxs={groupxs}')
        >>> groupxs = _compute_time_window(None, num_frames, heatmap_dates)
        >>> print(f'groupxs={groupxs}')
    """
    import kwarray
    from kwutil import util_time
    import numpy as np
    if window is None:
        bucket_ids = np.arange(num_frames)
    elif isinstance(window, str):
        assert heatmap_dates is not None
        delta = util_time.coerce_timedelta(window).total_seconds()
        image_unixtimes = np.array([d.timestamp() for d in heatmap_dates])
        image_unixtimes = image_unixtimes - image_unixtimes[0]
        bucket_ids = (image_unixtimes // delta).astype(int)
    elif isinstance(window, int):
        assert num_frames is not None
        frame_indexes = np.arange(num_frames)
        bucket_ids = frame_indexes // window
    else:
        raise NotImplementedError('')
    unique_ids, groupxs = kwarray.group_indices(bucket_ids)
    return groupxs


def _process_1_step(heatmaps, _agg_fn, track_bounds, step_idx, config):
    """
    Args:
        heatmaps (ndarray):
        _agg_fn (Callable):
        track_bounds (None | Coercable[kwimage.MultiPolygon]):
        step_idx (int):
        config (DataConfig):

    Returns:
        List[kwimage.Polygon]
    """
    # FIXME: no dynamic globals.
    if config.viz_out_dir is not None:
        viz_dpath = (config.viz_out_dir / f'heatmaps_{step_idx}').ensuredir()
        # print('\nviz_dpath = {}\n'.format(ub.urepr(viz_dpath, nl=1)))
    else:
        viz_dpath = None

    aggregated = _agg_fn(heatmaps,
                         thresh=config.thresh,
                         morph_kernel=config.morph_kernel,
                         norm_ord=config.norm_ord,
                         viz_dpath=viz_dpath)
    polygons = list(
        mask_to_polygons(aggregated,
                         thresh=config.thresh,
                         bounds=track_bounds,
                         thresh_hysteresis=config.thresh_hysteresis))
    return polygons


def _merge_polys(p1, t1, p2, t2, poly_merge_method=None):
    """
    Given two lists of polygons, p1 and p2, merge these according to:
      - add all unique polygons in the merged list
      - for overlapping polygons, add the union of both polygons

    Args:
        p1 (List[shapely.geometry.polygon.Polygon]):
            List of polygons in group1

        t1 (List[float]):
            List of times corresponding with polygons in group1

        p2 (List[shapely.geometry.polygon.Polygon]):
            List of polygons in group1

        t2 (List[float]):
            List of times corresponding with polygons in group2

        poly_merge_method (str):
            Codename for the algorithm used. Can be "v1", "v2", "v3", or "v3_noop".

    Example:
        >>> from geowatch.tasks.tracking.old_polygon_extraction import * # NOQA
        >>> from geowatch.tasks.tracking.old_polygon_extraction import _merge_polys  # NOQA
        >>> import kwimage
        >>> import numpy as np
        >>> #
        >>> p1 = [kwimage.Polygon.random().scale(0.2).to_shapely() for _ in range(1)]
        >>> t1 = np.arange(len(p1) * 2).reshape(-1, 2)
        >>> p2 = [kwimage.Polygon.random().to_shapely() for _ in range(1)]
        >>> t2 = np.arange(len(p2) * 2).reshape(-1, 2)
        >>> poly_merge_method = 'v3'
        >>> #
        >>> _merge_polys(p1, t1, p2, t2, poly_merge_method)

    Ignore:
        from geowatch.tasks.tracking.old_polygon_extraction import * # NOQA
        from geowatch.tasks.tracking.old_polygon_extraction import _merge_polys  # NOQA

        p1 = [kwimage.Polygon.random().scale(0.2).to_shapely() for _ in range(1)]
        p2 = [kwimage.Polygon.random().to_shapely() for _ in range(1)]
        unary_union(p1 + p2)
        _merge_polys(p1, p2)

        p1_kw = kwimage.Polygon(exterior=np.array([(0, 0), (1, 0), (0.5, 1)]))
        p2_kw = kwimage.Polygon(exterior=np.array([(0, 2), (1, 2), (0.5, 1)]))
        _p1 = p2_kw.to_shapely()
        _p2 = p1_kw.to_shapely()
        print(_p1.intersects(_p2))
        print(_p1.overlaps(_p2))
        print(unary_union([_p1, _p2]))

        p1_kw = kwimage.Boxes([[0, 0, 10, 10]], 'xywh').to_polygons()[0]
        p2_kw = kwimage.Boxes([[10, 0, 10, 10]], 'xywh').to_polygons()[0]
        _p1 = p2_kw.to_shapely()
        _p2 = p1_kw.to_shapely()
        print(_p1.intersects(_p2))
        print(_p1.overlaps(_p2))
        print(unary_union([_p1, _p2]))

        while True:
            _p1 = kwimage.Polygon.random().to_shapely()
            _p2 = kwimage.Polygon.random().to_shapely()
            if 1 or _p1.intersects(_p2):
                combo = unary_union([_p1, _p2])
                if combo.geom_type != 'Polygon':
                    raise Exception('!')
    """
    import numpy as np
    from shapely.ops import unary_union
    merged_polys = []
    merged_times = []

    if poly_merge_method is None:
        poly_merge_method = 'v1'

    elif poly_merge_method == 'v3_noop':
        merged_polys = p1 + p2
        merged_times = t1 + t2

    elif poly_merge_method == 'v3':
        p1_seen = set()
        p2_seen = set()
        # add all polygons that overlap
        for j, (_p1, _t1) in enumerate(zip(p1, t1)):
            if j in p1_seen:
                continue
            for i, (_p2, _t2) in enumerate(zip(p2, t2)):

                # if timestamps dont line up, skip
                if _t1[1] != _t2[0]:
                    continue
                if (i in p2_seen) or (i > len(p2) - 1):
                    continue

                if _p1.intersects(_p2):
                    combo = unary_union([_p1, _p2])
                    if combo.geom_type == 'Polygon':
                        merged_polys.append(combo)
                    elif combo.geom_type == 'MultiPolygon':
                        # Can this ever happen? It seems to have occurred in a test
                        # run. Bowties can cause this.
                        # import warnings
                        # warnings.warn('Found two intersecting polygons where the
                        # union was a multipolygon')
                        merged_polys.extend(list(combo.geoms))
                    else:
                        raise AssertionError(
                            f'Unexpected type {combo.geom_type} from {_p1} and {_p2}')

                    p1_seen.add(j)
                    p2_seen.add(i)

        # all polygons that did not overlap with any polygon
        all_p1 = set(np.arange(len(p1)))
        remaining_p1 = all_p1 - p1_seen

        for index in remaining_p1:
            merged_polys.append(p1[index])
            merged_times.append(t1[index])

        all_p2 = set(np.arange(len(p2)))
        remaining_p2 = all_p2 - p2_seen
        for index in remaining_p2:
            merged_polys.append(p2[index])
            merged_times.append(t2[index])

    elif poly_merge_method == 'v2':
        # Just combine anything that touches in both frames together
        from kwgis.utils import util_gis
        import geopandas as gpd
        geom_df = gpd.GeoDataFrame(geometry=p1 + p2)
        isect_idxs = util_gis.geopandas_pairwise_overlaps(geom_df, geom_df)
        level_sets = {frozenset(v.tolist()) for v in isect_idxs.values()}
        level_sets = list(map(sorted, level_sets))

        merged_polys = []
        for idxs in level_sets:
            if len(idxs) == 1:
                combo = geom_df['geometry'].iloc[idxs[0]]
                merged_polys.append(combo)
            else:
                combo = geom_df['geometry'].iloc[idxs].unary_union
                if combo.geom_type == 'Polygon':
                    merged_polys.append(combo)
                elif combo.geom_type == 'MultiPolygon':
                    # Can this ever happen? It seems to have occurred in a test
                    # run. Bowties can cause this.
                    # import warnings
                    # warnings.warn('Found two intersecting polygons where the
                    # union was a multipolygon')
                    # combo = combo.buffer(0)
                    merged_polys.append(combo.convex_hull)
                    # merged_polys.extend(list(combo.geoms))
                else:
                    raise AssertionError(f'Unexpected type {combo.geom_type}')

    elif poly_merge_method == 'v1':
        p1_seen = set()
        p2_seen = set()
        # add all polygons that overlap
        for j, _p1 in enumerate(p1):
            if j in p1_seen:
                continue
            for i, _p2 in enumerate(p2):
                if (i in p2_seen) or (i > len(p2) - 1):
                    continue
                if _p1.intersects(_p2):
                    combo = unary_union([_p1, _p2])
                    if combo.geom_type == 'Polygon':
                        merged_polys.append(combo)
                    elif combo.geom_type == 'MultiPolygon':
                        # Can this ever happen? It seems to have occurred in a test
                        # run. Bowties can cause this.
                        # import warnings
                        # warnings.warn('Found two intersecting polygons where the
                        # union was a multipolygon')
                        merged_polys.extend(list(combo.geoms))
                    else:
                        raise AssertionError(
                            f'Unexpected type {combo.geom_type} from {_p1} and {_p2}')

                    p1_seen.add(j)
                    p2_seen.add(i)

        # all polygons that did not overlap with any polygon
        all_p1 = set(np.arange(len(p1)))
        remaining_p1 = all_p1 - p1_seen

        for index in remaining_p1:
            merged_polys.append(p1[index])

        all_p2 = set(np.arange(len(p2)))
        remaining_p2 = all_p2 - p2_seen
        for index in remaining_p2:
            merged_polys.append(p2[index])
    else:
        raise ValueError(poly_merge_method)

    return merged_polys, merged_times



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class RunAllPointgenCLI(scfg.DataConfig):
    method = scfg.Value("ellipse", help="pointgen method")
    ignore_buffer = scfg.Value("20@10GSD", help="kwcoco ignore buffer size")
    size_prior = scfg.Value("20.06063 x 20.0141229 @ 10mGSD")
    time_prior = scfg.Value("1year")


def main(cmdline=1, **kwargs):
    """
    Developer script to get point-based polygons generated.
    Needs cleanup.

    Ignore:
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase3_data' --hardware=auto)
        DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase3_expt' --hardware=auto)
        echo "$DVC_DATA_DPATH"
        echo "$DVC_EXPT_DPATH"
    """
    import rich
    from rich.markup import escape

    config = RunAllPointgenCLI.cli(cmdline=1, strict=True)
    rich.print("config = " + escape(ub.urepr(config, nl=1)))

    import cmd_queue
    import geowatch

    data_dvc_dpath = geowatch.find_dvc_dpath(tags="phase3_data", hardware="ssd")
    # expt_dvc_dpath = geowatch.find_dvc_dpath(tags='phase3_expt', hardware='auto')

    kwcoco_bundle_dpath = data_dvc_dpath / "Drop8-ARA-Median10GSD-V1"

    points_fpath = data_dvc_dpath / "annotations/point_based_annotations.zip"
    region_dpath = (data_dvc_dpath / "annotations") / "drop8-v1/region_models"
    empty_region_dpath = (
        data_dvc_dpath / "annotations"
    ) / "drop8-v1/empty_region_models"

    if 0:
        # Based on T&E submodule
        points_fpath = (
            data_dvc_dpath
            / "submodules/annotations/supplemental_data/point_based_annotations.geojson"
        )
        region_dpath = data_dvc_dpath / "submodules/annotations/region_models/"
        empty_region_dpath = (
            data_dvc_dpath / "submodules/annotations/empty_region_models/"
        )

    assert points_fpath.exists()

    import geopandas as gpd

    # Read json text directly from the zipfile
    if points_fpath.endswith(".zip"):
        file = ub.zopen(points_fpath + "/" + "point_based_annotations.geojson", "r")
    else:
        file = open(points_fpath, "r")
    with file:
        points_gdf_crs84 = gpd.read_file(file)

    points_gdf_crs84["region_id"] = points_gdf_crs84["site_id"].apply(
        lambda s: "_".join(s.split("_")[0:2])
    )
    regions_with_points = sorted(points_gdf_crs84["region_id"].unique())

    common_params = dict(
        size_prior=config.size_prior,
        time_prior=config.time_prior,
        filepath_to_points=points_fpath,
    )

    # TODO: do a hash of the config in this output, or just specify it in config?
    dest_annot_dpath = (
        data_dvc_dpath / "annotations"
    ) / f"drop8-{config.method}-points-v1"

    dest_region_dpath = (dest_annot_dpath / "region_models").ensuredir()
    viz_region_dpath = (dest_annot_dpath / "region_models_viz").ensuredir()

    region_model_list = list(region_dpath.glob("*.geojson"))

    if 0:
        # Hack to add special point-only regions
        hacked_regions = ["HK_C001", "HK_C002"]
        for r in hacked_regions:
            region_fpath = empty_region_dpath / (r + ".geojson")
            assert region_fpath.exists()
            region_model_list.append(region_fpath)

    # Handle sites with "xxx" patterns that may be associated with multiple regions
    import kwutil

    region_with_points_patterns = []
    for region_id in regions_with_points:
        if region_id.endswith("xxx"):
            region_id = region_id.replace("xxx", "*")
        region_with_points_patterns.append(region_id)
    region_with_points_pattern = kwutil.util_pattern.MultiPattern.coerce(
        region_with_points_patterns
    )

    status_rows = []
    for region_path in region_model_list:
        region_id = region_path.stem.strip()
        kwcoco_path = (
            kwcoco_bundle_dpath
            / region_id
            / (f"imgonly-{region_id}-rawbands.kwcoco.zip")
        )
        # has_points = region_id in regions_with_points
        has_points = region_with_points_pattern.match(region_id)
        has_kwcoco = kwcoco_path.exists()
        row = {
            "region_id": region_id,
            "annotation_type": region_id.split("_")[1][0],
            "has_points": has_points,
            "has_kwcoco": has_kwcoco,
            "kwcoco_path": kwcoco_path,
        }
        status_rows.append(row)

    import pandas as pd
    import rich

    status_df = pd.DataFrame(status_rows)
    status_df.value_counts("has_kwcoco")
    # rows_to_process = status_df[status_df.has_kwcoco & status_df.has_points]
    # rows_to_process['annotation_type'].value_counts()
    # status_df[~status_df['has_kwcoco']]
    rich.print(status_df)

    polygen_template = ub.codeblock(
        r"""
        python -m geowatch.tasks.poly_from_point.predict \
            --method '{method}' \
            --region_id "{region_id}" \
            --filepath_to_images "{filepath_to_images}" \
            --filepath_to_points "{filepath_to_points}" \
            --filepath_to_region "{filepath_to_region}" \
            --time_prior "{time_prior}" \
            --size_prior "{size_prior}" \
            --filepath_output "{filepath_output}"
        """
    )

    regionviz_template = ub.codeblock(
        """
        geowatch draw_region "{region_fpath}" --fpath "{viz_fpath}"
        """
    )
    force_rerun = 1

    queue = cmd_queue.Queue.create(
        #backend="tmux",
        backend="serial",
        #size=16,
    )
    # queue.add_header_command(
    #     ub.codeblock(
    #         """
    #         pyenv shell 3.10.5
    #         source $(pyenv prefix)/envs/pyenv-geowatch/bin/activate
    #         """
    #     )
    # )
    for row in status_rows:
        region_id = row["region_id"]
        has_kwcoco = row["has_kwcoco"]
        has_points = row["has_points"]
        kwcoco_path = row["kwcoco_path"]

        filepath_output = dest_region_dpath / f"{region_id}.geojson"

        if has_kwcoco and has_points:
            fmtkw = {
                "method": config.method,
                "region_id": region_id,
                "filepath_to_images": None,
                "filepath_to_region": region_path,
                "filepath_output": filepath_output,
                **common_params,
            }
            polygen_cmd = polygen_template.format(**fmtkw)

            if not filepath_output.exists() or force_rerun:
                polygen_job = queue.submit(polygen_cmd, name=f"PolyGen_{region_id}")
            else:
                polygen_job = None

            DO_DRAW = 1
            if DO_DRAW:
                viz_fpath = viz_region_dpath / f"point_viz_{region_id}.png"
                if not viz_fpath.exists() or force_rerun:
                    regionviz_cmd = regionviz_template.format(
                        region_fpath=filepath_output,
                        viz_fpath=viz_fpath,
                    )
                    queue.submit(
                        regionviz_cmd, name=f"Viz_{region_id}", depends=polygen_job
                    )

            final_out_kwcoco_fpath = (
                kwcoco_bundle_dpath
                / region_id
                / (f"pointannv4_ellipse-{region_id}-rawbands.kwcoco.zip")
            )

            if config.ignore_buffer:
                reproject_out_kwcoco_fpath = (
                    kwcoco_bundle_dpath
                    / region_id
                    / (f"_preignore-pointannv4_ellipse-{region_id}-rawbands.kwcoco.zip")
                )
            else:
                reproject_out_kwcoco_fpath = final_out_kwcoco_fpath

            reproject_cmd = ub.codeblock(
                rf"""
                geowatch reproject_annotations \
                    --src "{kwcoco_path}" \
                    --dst "{reproject_out_kwcoco_fpath}" \
                    --region_models="{filepath_output}"
                """
            )
            if not reproject_out_kwcoco_fpath.exists() or force_rerun:
                reproject_job = queue.submit(
                    reproject_cmd, name=f"reproject_{region_id}", depends=polygen_job
                )
            else:
                reproject_job = None

            if config.ignore_buffer:
                ignore_buffer_kwcoco_fpath = final_out_kwcoco_fpath
                add_kwcoco_ignore_buffer_command = ub.codeblock(
                    rf"""
                    python -m geowatch.cli.coco_add_ignore_buffer \
                        --ignore_buffer_size '{config.ignore_buffer}' \
                        --src "{reproject_out_kwcoco_fpath}" \
                        --dst "{ignore_buffer_kwcoco_fpath}"
                    """
                )
                if not ignore_buffer_kwcoco_fpath.exists() or force_rerun:
                    add_ignore_buffer_job = queue.submit(
                        add_kwcoco_ignore_buffer_command,
                        name=f"add_ignore_buffer_{region_id}",
                        depends=reproject_job,
                    )
                else:
                    add_ignore_buffer_job = None  # NOQA

    queue.print_graph()
    queue.print_commands()
    #queue.run()

    print(
        ub.codeblock(
            r"""
        Next Step is to run this splitgen:

        export DST_DVC_DATA_DPATH=$(geowatch_dvc --tags='phase3_data' --hardware=ssd)
        export DST_BUNDLE_DPATH=$DST_DVC_DATA_DPATH/Drop8-ARA-Median10GSD-V1

        python -m geowatch.cli.queue_cli.prepare_splits \
            --src_kwcocos "$DST_BUNDLE_DPATH"/*/pointannv4_ellipse-*-rawbands.kwcoco.zip \
            --dst_dpath "$DST_BUNDLE_DPATH" \
            --suffix=rawbands_pointannv4_ellipse \
            --backend=tmux --tmux_workers=2 \
            --splits split6 \
            --run=1
        """
        )
    )


if __name__ == "__main__":
    """
    CommandLine:
        python -m geowatch.tasks.poly_from_point.run_all_pointgen
    """
    main()



#!/usr/bin/env python3
"""
SeeAlso:
    ~/code/watch/geowatch/cli/coco_time_combine.py
    ~/code/watch/geowatch/cli/queue_cli/prepare_teamfeats.py
"""
import scriptconfig as scfg
import ubelt as ub
from cmd_queue.cli_boilerplate import CMDQueueConfig


class PrepareTimeAverages(CMDQueueConfig):
    """
    Prepare a temporally averaged dataset on multiple regions
    """
    regions = scfg.Value('all', type=str, help=ub.paragraph(
        '''
        The regions to time average (this is not a robust implementation)
        '''))

    resolution = '10GSD'

    time_window = '1year'
    merge_method = 'mean'
    remove_seasons = scfg.Value([], nargs='+')
    spatial_tile_size = None
    mask_low_quality = scfg.Value(True)

    combine_workers = scfg.Value(4, help='number of workers per combine job')

    input_bundle_dpath = scfg.Value(None)
    output_bundle_dpath = scfg.Value(None)

    skip_existing = scfg.Value(True)
    cache = scfg.Value(True)

    reproject = scfg.Value(False, isflag=True, help='Enable reprojection of annotations. Requires true_site_dpath and true_region_dpath be specified')
    true_site_dpath = scfg.Value(None)
    true_region_dpath = scfg.Value(None)

    max_images_per_group = scfg.Value(None, help=ub.paragraph(
        '''
        If specified, the averaging operation (e.g. mean/median) will only
        consider a subset of the images within each temporal window.  This can
        greatly reduce the resources required to run this script at the cost of
        quality. Currently a heuristic is used to select the "highest quality"
        subset of images.
        '''))

    queue_name = scfg.Value('time-ave-queue', help='overwrite the default queue name', group='cmd-queue')


def _find_valid_regions():
    import geowatch
    dvc_data_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    coco_fpaths = list((dvc_data_dpath / 'Drop6').glob('imgonly*.kwcoco*'))
    import kwcoco
    dsets = list(kwcoco.CocoDataset.coerce_multiple(coco_fpaths, workers='avail'))

    for dset in dsets:
        if dset.n_images > 0:
            print('- ' + ub.Path(dset.fpath).name.split('-')[1].split('.')[0])


def main(cmdline=1, **kwargs):
    """
    Example:
        >>> # xdoctest: +SKIP
        >>> cmdline = 0
        >>> kwargs = dict(
        >>> )
        >>> main(cmdline=cmdline, **kwargs)
    """
    config = PrepareTimeAverages.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    from kwutil.partial_format import subtemplate
    rich.print('config = ' + ub.urepr(config, nl=1))
    assert config.output_bundle_dpath is not None
    assert config.input_bundle_dpath is not None
    # import geowatch
    # dvc_data_dpath = geowatch.find_dvc_dpath(tags='phase2_data', hardware='auto')

    if config.regions == 'all':
        all_regions = [p.name.split('.')[0] for p in (ub.Path(config.true_region_dpath)).ls()]
        chosen_regions = all_regions
    elif config.regions == 'all_tne':
        all_regions = [p.name.split('.')[0] for p in (ub.Path(config.true_region_dpath)).ls()]
        tne_regions = [r for r in all_regions if r.split('_')[1].startswith('R')]
        chosen_regions = tne_regions
    else:
        from kwutil.util_yaml import Yaml
        chosen_regions = Yaml.coerce(config.regions)

    from geowatch.mlops.pipeline_nodes import ProcessNode

    rich.print('chosen_regions = {}'.format(ub.urepr(chosen_regions, nl=1)))

    # time_duration = '1year'
    # time_duration = '3months'
    # all_regions = [
    #     'KR_R001',
    #     'KR_R002',
    #     'NZ_R001',
    #     'CH_R001',
    #     'BR_R001',
    #     'BR_R002',
    #     'BH_R001',
    #     'AE_R001',
    # ]

    queue = config.create_queue()

    # Need these for landcover
    other_s2_bands = '|coastal|cirrus|B05|B06|B07|B8A|B09'

    def submit_job_step(node, depends=None):
        if config.skip_existing and node.outputs_exist:
            job = None
        else:
            node.cache = config.cache
            job = queue.submit(node.final_command(), depends=depends, name=node.name)
        return job

    for region in chosen_regions:

        fmtdict = dict(
            # DVC_DATA_DPATH=dvc_data_dpath,
            INPUT_BUNDLE_DPATH=config.input_bundle_dpath,
            OUTPUT_BUNDLE_DPATH=config.output_bundle_dpath,
            REGION=region,
            # SUFFIX='MeanYear',
            # TIME_DURATION='3months',
            # SUFFIX='Mean3Month10GSD',
            TRUE_SITE_DPATH=config.true_site_dpath,
            TRUE_REGION_DPATH=config.true_region_dpath,
            CHANNELS='red|green|blue|nir|swir16|swir22|pan' + other_s2_bands,
            remove_seasons_str=None if not config.remove_seasons else ','.join(config.remove_seasons),
            mask_low_quality=config.mask_low_quality,
        )
        fmtdict.update(config)

        input_bundle_dpath = ub.Path(config.input_bundle_dpath)
        output_bundle_dpath = ub.Path(config.output_bundle_dpath)

        INPUT_KWCOCO_FPATH = input_bundle_dpath / region / f'imgonly-{region}-rawbands.kwcoco.zip'
        TAVE_KWCOCO_FPATH = output_bundle_dpath / region / f'_unfielded_imgonly-{region}-rawbands.kwcoco.zip'
        FIELDED_KWCOCO_FPATH = output_bundle_dpath / region / f'imgonly-{region}-rawbands.kwcoco.zip'
        FINAL_KWCOCO_FPATH = output_bundle_dpath / region / f'imganns-{region}-rawbands.kwcoco.zip'

        fmtdict['INPUT_KWCOCO_FPATH'] = INPUT_KWCOCO_FPATH
        fmtdict['TAVE_KWCOCO_FPATH'] = TAVE_KWCOCO_FPATH
        fmtdict['FIELDED_KWCOCO_FPATH'] = FIELDED_KWCOCO_FPATH
        fmtdict['FINAL_KWCOCO_FPATH'] = FINAL_KWCOCO_FPATH

        code = subtemplate(ub.codeblock(
            r'''
            python -m geowatch.cli.coco_time_combine \
                --kwcoco_fpath="$INPUT_KWCOCO_FPATH" \
                --output_kwcoco_fpath="$TAVE_KWCOCO_FPATH" \
                --channels="$CHANNELS" \
                --resolution="$resolution" \
                --time_window=$time_window \
                --remove_seasons=$remove_seasons_str \
                --merge_method=$merge_method \
                --spatial_tile_size=$spatial_tile_size \
                --mask_low_quality=$mask_low_quality \
                --max_images_per_group=$max_images_per_group \
                --start_time=2010-03-01 \
                --assets_dname="raw_bands" \
                --workers=$combine_workers
            '''), fmtdict)
        node = ProcessNode(
            name=f'combine-time-{region}',
            command=code,
            in_paths={'kwcoco_fpath': subtemplate('$INPUT_KWCOCO_FPATH', fmtdict)},
            out_paths={'output_kwcoco_fpath': subtemplate('$TAVE_KWCOCO_FPATH', fmtdict)},
            _no_outarg=True,
            _no_inarg=True,
        )
        combine_job = submit_job_step(node)

        code = subtemplate(ub.codeblock(
            r'''
            python -m geowatch add_fields \
                --src $TAVE_KWCOCO_FPATH \
                --dst $FIELDED_KWCOCO_FPATH
            '''), fmtdict)
        node = ProcessNode(
            name=f'add-fields-{region}',
            command=code,
            in_paths={'src': subtemplate('$TAVE_KWCOCO_FPATH', fmtdict)},
            out_paths={'dst': subtemplate('$FIELDED_KWCOCO_FPATH', fmtdict)},
            _no_outarg=True,
            _no_inarg=True,
        )
        field_job = submit_job_step(node, depends=[combine_job])

        if config.reproject:
            code = subtemplate(ub.codeblock(
                r'''
                python -m geowatch reproject \
                    --src $FIELDED_KWCOCO_FPATH \
                    --dst $FINAL_KWCOCO_FPATH \
                    --status_to_catname="positive_excluded: positive" \
                    --regions="$TRUE_REGION_DPATH/${REGION}.geojson" \
                    --sites="$TRUE_SITE_DPATH/${REGION}_*.geojson"
                '''), fmtdict)
            node = ProcessNode(
                command=code,
                name=f'reproject-ann-{region}',
                in_paths={'src': subtemplate('$FIELDED_KWCOCO_FPATH', fmtdict)},
                out_paths={'dst': subtemplate('$FINAL_KWCOCO_FPATH', fmtdict)},
                _no_outarg=True,
                _no_inarg=True,
            )
            field_job = submit_job_step(node, depends=[field_job])

    config.run_queue(queue)


SUMMER_CONFIG = """

DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=hdd)
SSD_DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=ssd)
python -m geowatch.cli.queue_cli.prepare_time_combined_dataset \
    --regions="[
            # T&E Regions
            AE_R001, BH_R001, BR_R001, BR_R002, BR_R004, BR_R005, CH_R001,
            KR_R001,
            KR_R002, LT_R001, NZ_R001, US_R001, US_R004, US_R005,
            US_R006, US_R007,
            # iMerit Regions
            AE_C001,
            AE_C002,
            AE_C003, PE_C001, QA_C001, SA_C005, US_C000, US_C010,
            US_C011, US_C012,
    ]" \
    --input_bundle_dpath=$SSD_DVC_DATA_DPATH/Drop6 \
    --output_bundle_dpath=$DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD \
    --true_site_dpath=$DVC_DATA_DPATH/annotations/drop6_hard_v1/site_models \
    --true_region_dpath=$DVC_DATA_DPATH/annotations/drop6_hard_v1/region_models \
    --spatial_tile_size=256 \
    --merge_method=median \
    --remove_seasons=winter \
    --tmux_workers=2 \
    --time_window=1y \
    --combine_workers=4 \
    --resolution=10GSD \
    --backend=tmux \
    --run=1

# Drop 6
export CUDA_VISIBLE_DEVICES="0,1"
DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware='auto')
BUNDLE_DPATH=$DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD
python -m geowatch.cli.queue_cli.prepare_teamfeats \
    --base_fpath "$BUNDLE_DPATH"/imganns-*[0-9].kwcoco.zip \
    --expt_dvc_dpath="$DVC_EXPT_DPATH" \
    --with_landcover=1 \
    --with_invariants2=1 \
    --with_materials=0 \
    --with_depth=0 \
    --with_cold=0 \
    --skip_existing=1 \
    --assets_dname=teamfeats \
    --gres=0,1 --tmux_workers=4 --backend=tmux --run=1

# DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
# python -m geowatch.cli.queue_cli.prepare_splits \
#     --base_fpath=$DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD/combo_imganns*_L*.kwcoco.zip \
#     --constructive_mode=True \
#     --suffix=L \
#     --backend=tmux --workers=6 \
#     --run=1

DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
python -m geowatch.cli.queue_cli.prepare_splits \
    --base_fpath=$DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD/combo_imganns*_I2L*.kwcoco.zip \
    --constructive_mode=True \
    --suffix=I2L \
    --backend=tmux --tmux_workers=6 \
    --run=1

# DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
# python -m geowatch.cli.queue_cli.prepare_splits \
#     --base_fpath=$DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD/imganns-*.kwcoco.zip \
#     --constructive_mode=True \
#     --suffix=rawbands \
#     --backend=tmux --tmux_workers=6 \
#     --run=1

DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware='auto')
TRUE_SITE_DPATH=$DVC_DATA_DPATH/annotations/drop6_hard_v1/site_models
OUTPUT_BUNDLE_DPATH=$DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD
geowatch reproject \
    --src $DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD/data_vali_I2L_split6.kwcoco.zip \
    --inplace=True \
    --status_to_catname="positive_excluded: positive" \
    --site_models=$TRUE_SITE_DPATH
geowatch reproject \
    --src $DVC_DATA_DPATH/Drop6-NoWinterMedian10GSD/data_train_I2L_split6.kwcoco.zip \
    --inplace=True \
    --status_to_catname="positive_excluded: positive" \
    --site_models=$TRUE_SITE_DPATH

"""


if __name__ == '__main__':
    r"""

    CommandLine:

        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        python geowatch.cli.queue_cli.prepare_time_combined_dataset.py \
            --regions="[
                    # T&E Regions
                    AE_R001, BH_R001, BR_R001, BR_R002, BR_R004, BR_R005, CH_R001,
                    KR_R001, KR_R002, LT_R001, NZ_R001, US_R001, US_R004, US_R005,
                    US_R006, US_R007,
                    # iMerit Regions
                    AE_C001,
                    AE_C002,
                    AE_C003, PE_C001, QA_C001, SA_C005, US_C000, US_C010,
                    US_C011, US_C012,
            ]" \
            --input_bundle_dpath=$DVC_DATA_DPATH/Drop6 \
            --output_bundle_dpath=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2 \
            --true_site_dpath=$DVC_DATA_DPATH/annotations/drop6_hard_v1/site_models \
            --true_region_dpath=$DVC_DATA_DPATH/annotations/drop6_hard_v1/region_models \
            --backend=tmux \
            --tmux_workers=4 \
            --combine_workers=2 \
            --resolution=10GSD \
            --run=1

        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        python -m geowatch.cli.queue_cli.prepare_splits \
            --base_fpath=$DVC_DATA_DPATH/Drop6-MeanYear10GSD/imganns-*.kwcoco.zip \
            --constructive_mode=True \
            --suffix=rawbands \
            --backend=tmux --tmux_workers=6 \
            --run=1

        geowatch visualize /home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Drop6-MeanYear10GSD/imganns-NZ_R001.kwcoco.zip --smart

        # Drop 6
        export CUDA_VISIBLE_DEVICES="0,1"
        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware='auto')
        BUNDLE_DPATH=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2
        python -m geowatch.cli.queue_cli.prepare_teamfeats \
            --base_fpath "$BUNDLE_DPATH"/imganns-*[0-9].kwcoco.zip \
            --expt_dvc_dpath="$DVC_EXPT_DPATH" \
            --with_landcover=1 \
            --with_invariants2=1 \
            --with_materials=0 \
            --with_depth=0 \
            --with_cold=0 \
            --skip_existing=1 \
            --assets_dname=teamfeats \
            --gres=0,1 --tmux_workers=4 --backend=tmux --run=0

        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        python -m geowatch.cli.queue_cli.prepare_splits \
            --base_fpath=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/combo_imganns*_I2L*.kwcoco.zip \
            --constructive_mode=True \
            --suffix=I2L \
            --backend=tmux --tmux_workers=6 \
            --run=1

        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware='auto')
        TRUE_SITE_DPATH=$DVC_DATA_DPATH/annotations/drop6_hard_v1/site_models
        OUTPUT_BUNDLE_DPATH=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2
        python -m geowatch reproject \
            --src data_vali_I2L_split6.kwcoco.zip \
            --dst data_vali_I2L_split6.kwcoco.zip \
            --site_models=$TRUE_SITE_DPATH

        DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
        DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware='auto')
        TRUE_SITE_DPATH=$DVC_DATA_DPATH/annotations/drop6_hard_v1/site_models
        OUTPUT_BUNDLE_DPATH=$DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2
        python -m geowatch reproject \
            --src data_train_I2L_split6.kwcoco.zip \
            --dst data_train_I2L_split6.kwcoco.zip \
            --site_models=$TRUE_SITE_DPATH
    """
    main()



"""
DEPRECATED: moved to kwgis

Utilities for geopandas and other geographic information system tools
"""
from kwgis.utils.util_gis import plot_geo_background  # NOQA
from kwgis.utils.util_gis import geopandas_pairwise_overlaps  # NOQA
from kwgis.utils.util_gis import latlon_text  # NOQA
from kwgis.utils.util_gis import demo_regions_geojson_text  # NOQA
from kwgis.utils.util_gis import load_geojson  # NOQA
from kwgis.utils.util_gis import read_geojson  # NOQA
from kwgis.utils.util_gis import get_crs84  # NOQA
from kwgis.utils.util_gis import _get_crs84  # NOQA
from kwgis.utils.util_gis import _flip  # NOQA
from kwgis.utils.util_gis import shapely_flip_xy  # NOQA
from kwgis.utils.util_gis import project_gdf_to_local_utm  # NOQA
from kwgis.utils.util_gis import UTM_TransformContext  # NOQA
from kwgis.utils.util_gis import _demo_convert_latlon_to_utm  # NOQA
from kwgis.utils.util_gis import find_local_meter_epsg_crs  # NOQA
from kwgis.utils.util_gis import check_latlons  # NOQA
from kwgis.utils.util_gis import coerce_geojson_datas  # NOQA
from kwgis.utils.util_gis import coerce_geojson_paths  # NOQA
from kwgis.utils.util_gis import _coerce_raw_geojson  # NOQA
from kwgis.utils.util_gis import _load_json_from_path  # NOQA
from kwgis.utils.util_gis import load_geojson_datas  # NOQA
from kwgis.utils.util_gis import crs_geojson_to_gdf  # NOQA
from kwgis.utils.util_gis import coerce_crs  # NOQA
from kwgis.utils.util_gis import utm_epsg_from_latlon  # NOQA


def utm_epsg_from_zoneletter(zone, letter):
    """
    Example:
        >>> lon, lat = -99.1149386165299831, 19.4867151641771414
        >>> espg1 = utm_epsg_from_latlon(lat, lon)
        >>> zone = 14
        >>> letter = 'Q'
        >>> espg2 = utm_epsg_from_zoneletter(zone, letter)
        >>> assert espg2 == espg1

    References:
        https://www.maptools.com/tutorials/grid_zone_details

    Notes:
        The UTM coordinate system divides the earth into 60 zones each 6
        degrees of longitude wide. These zones define the reference point for
        UTM grid coordinates within the zone. UTM zones extend from a latitude
        of 80° S to 84° N. In the polar regions the Universal Polar
        Stereographic (UPS) grid system is used. Note that there are a few
        exceptions to zone width in Northern Europe to keep small countries in
        a single zone.

        UTM zones are numbered 1 through 60, starting at the international date
        line, longitude 180°, and proceeding east. Zone 1 extends from 180° W
        to 174° W and is centered on 177° W.

        Each zone is divided into horizontal bands spanning 8 degrees of
        latitude. These bands are lettered, south to north, beginning at 80° S
        with the letter C and ending with the letter X at 84° N. The letters I
        and O are skipped to avoid confusion with the numbers one and zero. The
        band lettered X spans 12° of latitude.
    """
    if letter.upper() >= 'N':
        epsg_code = 32600 + int(zone)  # Northern hemisphere
    else:
        epsg_code = 32700 + int(zone)  # Southern hemisphere
    return epsg_code
